@@ -55,7 +55,30 @@ module Hashie
   #   mash.author # => <Mash>
   #
   class Mash < Hash
-    include Hashie::PrettyInspect
+    include Hashie::Extensions::PrettyInspect
+
+    ALLOWED_SUFFIXES = %w(? ! = _)
+    SUFFIXES_PARSER  = /(.*?)([#{ALLOWED_SUFFIXES.join}]?)$/
+
+    def self.load(path, options = {})
+      @_mashes ||= new
+
+      return @_mashes[path] if @_mashes.key?(path)
+      fail ArgumentError, "The following file doesn't exist: #{path}" unless File.file?(path)
+
+      parser = options.fetch(:parser) {  Hashie::Extensions::Parsers::YamlErbParser }
+      @_mashes[path] = new(parser.perform(path)).freeze
+    end
+
+    def to_module(mash_method_name = :settings)
+      mash = self
+      Module.new do |m|
+        m.send :define_method, mash_method_name.to_sym do
+          mash
+        end
+      end
+    end
+
     alias_method :to_s, :inspect
 
     # If you pass in an existing hash, it will
@@ -67,22 +90,15 @@ module Hashie
       default ? super(default) : super(&blk)
     end
 
-    class << self; alias [] new; end
-
-    def id #:nodoc:
-      self["id"]
-    end
-
-    def type #:nodoc:
-      self["type"]
-    end
+    class << self; alias_method :[], :new; end
 
     alias_method :regular_reader, :[]
     alias_method :regular_writer, :[]=
 
     # Retrieves an attribute set in the Mash. Will convert
     # any key passed in to a string before retrieving.
-    def [](key)
+    def custom_reader(key)
+      default_proc.call(self, key) if default_proc && !key?(key)
       value = regular_reader(convert_key(key))
       yield value if block_given?
       value
@@ -91,10 +107,13 @@ module Hashie
     # Sets an attribute in the Mash. Key will be converted to
     # a string before it is set, and Hashes will be converted
     # into Mashes for nesting purposes.
-    def []=(key,value) #:nodoc:
-      regular_writer(convert_key(key), convert_value(value))
+    def custom_writer(key, value, convert = true) #:nodoc:
+      regular_writer(convert_key(key), convert ? convert_value(value) : value)
     end
 
+    alias_method :[], :custom_reader
+    alias_method :[]=, :custom_writer
+
     # This is the bang method reader, it will return a new Mash
     # if there isn't a value already assigned to the key requested.
     def initializing_reader(key)
@@ -114,18 +133,22 @@ module Hashie
       end
     end
 
-    def fetch(key, default_value = nil)
-      self[key] || block_given? && yield(key) || default_value || super(key)
+    def fetch(key, *args)
+      super(convert_key(key), *args)
     end
 
     def delete(key)
       super(convert_key(key))
     end
 
+    def values_at(*keys)
+      super(*keys.map { |key| convert_key(key) })
+    end
+
     alias_method :regular_dup, :dup
     # Duplicates the current mash as a new mash.
     def dup
-      self.class.new(self, self.default)
+      self.class.new(self, default)
     end
 
     def key?(key)
@@ -145,14 +168,14 @@ module Hashie
     # Recursively merges this mash with the passed
     # in hash, merging each hash in the hierarchy.
     def deep_update(other_hash, &blk)
-      other_hash.each_pair do |k,v|
+      other_hash.each_pair do |k, v|
         key = convert_key(k)
-        if regular_reader(key).is_a?(Mash) and v.is_a?(::Hash)
-          regular_reader(key).deep_update(v, &blk)
+        if regular_reader(key).is_a?(Mash) && v.is_a?(::Hash)
+          custom_reader(key).deep_update(v, &blk)
         else
           value = convert_value(v, true)
-          value = blk.call(key, self[k], value) if blk
-          regular_writer(key, value)
+          value = convert_value(blk.call(key, self[k], value), true) if blk && self.key?(k)
+          custom_writer(key, value, false)
         end
       end
       self
@@ -161,6 +184,11 @@ module Hashie
     alias_method :update, :deep_update
     alias_method :merge!, :update
 
+    # Assigns a value to a key
+    def assign_property(name, value)
+      self[name] = value
+    end
+
     # Performs a shallow_update on a duplicate of the current mash
     def shallow_merge(other_hash)
       dup.shallow_update(other_hash)
@@ -169,7 +197,7 @@ module Hashie
     # Merges (non-recursively) the hash from the argument,
     # changing the receiving hash
     def shallow_update(other_hash)
-      other_hash.each_pair do |k,v|
+      other_hash.each_pair do |k, v|
         regular_writer(convert_key(k), convert_value(v, true))
       end
       self
@@ -181,49 +209,73 @@ module Hashie
       self
     end
 
-    # Will return true if the Mash has had a key
-    # set in addition to normal respond_to? functionality.
-    def respond_to?(method_name, include_private=false)
-      return true if key?(method_name) || method_name.to_s.slice(/[=?!_]\Z/)
-      super
+    def respond_to_missing?(method_name, *args)
+      return true if key?(method_name)
+      _, suffix = method_suffix(method_name)
+      case suffix
+      when '=', '?', '!', '_'
+        return true
+      else
+        super
+      end
+    end
+
+    def prefix_method?(method_name)
+      method_name = method_name.to_s
+      method_name.end_with?(*ALLOWED_SUFFIXES) && key?(method_name.chop)
     end
 
     def method_missing(method_name, *args, &blk)
       return self.[](method_name, &blk) if key?(method_name)
-      match = method_name.to_s.match(/(.*?)([?=!_]?)$/)
-      case match[2]
-      when "="
-        self[match[1]] = args.first
-      when "?"
-        !!self[match[1]]
-      when "!"
-        initializing_reader(match[1])
-      when "_"
-        underbang_reader(match[1])
+      name, suffix = method_suffix(method_name)
+      case suffix
+      when '='
+        assign_property(name, args.first)
+      when '?'
+        !!self[name]
+      when '!'
+        initializing_reader(name)
+      when '_'
+        underbang_reader(name)
       else
-        default(method_name, *args, &blk)
+        self[method_name]
       end
     end
 
+    # play nice with ActiveSupport Array#extract_options!
+    def extractable_options?
+      true
+    end
+
+    # another ActiveSupport method, see issue #270
+    def reverse_merge(other_hash)
+      Hashie::Mash.new(other_hash).merge(self)
+    end
+
     protected
 
+    def method_suffix(method_name)
+      match = method_name.to_s.match(SUFFIXES_PARSER)
+      [match[1], match[2]]
+    end
+
     def convert_key(key) #:nodoc:
       key.to_s
     end
 
-    def convert_value(val, duping=false) #:nodoc:
+    def convert_value(val, duping = false) #:nodoc:
       case val
-        when self.class
-          val.dup
-        when Hash
-          duping ? val.dup : val
-        when ::Hash
-          val = val.dup if duping
-          self.class.new(val)
-        when Array
-          val.collect{ |e| convert_value(e) }
-        else
-          val
+      when self.class
+        val.dup
+      when Hash
+        duping ? val.dup : val
+      when ::Hash
+        val = val.dup if duping
+        self.class.new(val)
+      when Array
+        val.map { |e| convert_value(e) }
+      else
+        val
       end
     end
   end
